from typing import Any
from types import FunctionType, ModuleType, CellType
from serializers.base.constants import is_iterable
from serializers.base.constants import (PRIMITIVES, COLLECTIONS, CODE_ATTRIBUTES,
                                          TYPE, VALUE, ITERATOR_TYPE,
                                          IGNORED_CLASS_ATTRIBUTES, IGNORED_TYPES)
import inspect



class Serializer:

    @staticmethod
    def pack(obj: Any) -> dict:
        """
        Method that converts object to the dictionary
        that represents the object.

        Parameters:
        obj (Any): Any Python object.

        Returns:
        dictionary.
        """

        if isinstance(obj, PRIMITIVES):
            return {TYPE: type(obj).__name__,
                    VALUE: str(obj) if isinstance(obj, complex) else obj}

        if isinstance(obj, COLLECTIONS):
            if isinstance(obj, dict):
                return {TYPE: type(obj).__name__,
                        VALUE: [[Serializer.pack(key), Serializer.pack(value)]
                                for key, value in obj.items()]}
            else:
                return {TYPE: type(obj).__name__,
                        VALUE: [Serializer.pack(item) for item in obj]}

        if inspect.isfunction(obj):
            return {TYPE: type(obj).__name__,
                    VALUE: Serializer._pack_function(obj)}

        if inspect.iscode(obj):
            return {TYPE: type(obj).__name__,
                    VALUE: {key: Serializer.pack(value)
                            for key, value in inspect.getmembers(obj)
                            if key in CODE_ATTRIBUTES}}

        if isinstance(obj, CellType):
            return {TYPE: type(obj).__name__,
                    VALUE: Serializer.pack(obj.cell_contents)}

        if inspect.isclass(obj):
            return {TYPE: "class",
                    VALUE: Serializer._pack_class(obj)}

        if isinstance(obj, property):
            return {TYPE: type(obj).__name__,
                    VALUE: {"fget": Serializer.pack(obj.fget),
                            "fset": Serializer.pack(obj.fset),
                            "fdel": Serializer.pack(obj.fdel)}}

        if is_iterable(obj):
            return {TYPE: ITERATOR_TYPE,
                    VALUE: [Serializer.pack(item) for item in obj]}

        # if inspect.isgenerator(obj):
        #     raise TypeError("Python does not support creating generators on the fly")

        if inspect.ismethod(obj):
            return {TYPE: type(obj).__name__,
                    VALUE: Serializer._pack_function(obj.__func__)}

        return {TYPE: "object",
                VALUE: {"__class__": Serializer.pack(obj.__class__),
                        "__vars__": {key: Serializer.pack(value) for key, value in vars(obj).items()}}}

    @staticmethod
    def _pack_function(obj: FunctionType, cls=None):

        return {"__name__": obj.__name__,
                "__globals__": Serializer._pack_globals(obj, cls),
                "__closure__": Serializer.pack(obj.__closure__),
                "__defaults__": Serializer.pack(obj.__defaults__),
                "__kwdefaults__": Serializer.pack(obj.__kwdefaults__),
                "__code__": {key: Serializer.pack(value)
                             for key, value in inspect.getmembers(obj.__code__)
                             if key in CODE_ATTRIBUTES}}

    @staticmethod
    def _pack_globals(obj, cls=None):
        globs = dict()

        for key, value in obj.__globals__.items():

            if key not in obj.__code__.co_names:
                continue

            if isinstance(value, ModuleType):
                globs[f"module {key}"] = Serializer.pack(key)

            elif inspect.isclass(value):
                if cls and value != cls or not cls:
                    globs[key] = Serializer.pack(value)

            elif key == obj.__code__.co_name:
                globs[key] = Serializer.pack(obj.__name__)

            else:
                globs[key] = Serializer.pack(value)

        return globs

    @staticmethod
    def _pack_class(obj):
        packed = dict()
        packed["__name__"] = Serializer.pack(obj.__name__)

        for key, value in obj.__dict__.items():

            if key in IGNORED_CLASS_ATTRIBUTES or type(value) in IGNORED_TYPES:
                continue

            if isinstance(obj.__dict__[key], staticmethod):
                packed[key] = dict()
                packed[key]["type"] = "staticmethod"
                packed[key]["value"] = {"type": "function", "value": Serializer._pack_function(value.__func__, obj)}

            elif isinstance(obj.__dict__[key], classmethod):
                packed[key] = dict()
                packed[key]["type"] = "classmethod"
                packed[key]["value"] = {"type": "function", "value": Serializer._pack_function(value.__func__, obj)}

            elif inspect.ismethod(value):
                packed[key] = Serializer._pack_function(value.__func__, obj)

            elif inspect.isfunction(value):
                packed[key] = dict()
                packed[key]["type"] = "function"
                packed[key]["value"] = Serializer._pack_function(value, obj)

            else:
                packed[key] = Serializer.pack(value)

        packed["__bases__"] = dict()
        packed["__bases__"]["type"] = "tuple"
        packed["__bases__"]["value"] = [Serializer.pack(item) for item in obj.__bases__ if item != object]

        return packed







import unittest
from serializers.json_serializer.JsonSerializer import JsonSerializer
from serializers.xml_serializer.XmlSerializer import XmlSerializer
from tests.constants_for_testing import (circle_area, square_area, generator, bts, bts_arr, sum_func,
                        generator_expression, closure, factorial, Profile,
                        sum_args, sum_kwargs, sum_args_kwargs, subgenerator,
                        lambda_pow, C, E, Human, func1, func2, func3, it)



class BaseJsonTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.serializer = JsonSerializer()
        #self.serializer = XmlSerializer()
        self.alias = lambda x: self.serializer.loads(self.serializer.dumps(x))


class PrimitivesTestCase(BaseJsonTestCase):
    def test_int(self):
        self.assertEqual(10, self.alias(10))
        self.assertEqual(0, 0)
        self.assertEqual(-10, self.alias(-10))

    def test_float(self):
        self.assertEqual(10.1, self.alias(10.1))
        self.assertEqual(-10.1, self.alias(-10.1))

    def test_bool(self):
        self.assertEqual(True, self.alias(True))
        self.assertEqual(False, self.alias(False))

    def test_str(self):
        self.assertEqual("", self.alias(""))
        self.assertEqual("Hello world!", self.alias("Hello world!"))

    def test_none(self):
        self.assertEqual(None, self.alias(None))

    def test_complex(self):
        self.assertEqual(5 + 3j, self.alias(5 + 3j))
        self.assertEqual(5 + 0j, self.alias(5 + 0j))
        self.assertEqual(3j, self.alias(3j))
        self.assertEqual(0j, self.alias(0j))


class CollectionsTestCase(BaseJsonTestCase):
    def test_list(self):
        self.assertEqual([], self.alias([]))
        self.assertEqual([1, 2, 3], self.alias([1, 2, 3]))

    def test_tuple(self):
        self.assertEqual((), self.alias(()))
        self.assertEqual((1, 2, 3), self.alias((1, 2, 3)))

    def test_set(self):
        self.assertEqual({}, self.alias({}))
        self.assertEqual({1, 2, 3}, self.alias({1, 2, 3}))

    def test_dict(self):
        self.assertEqual({'a1': 'b2'}, self.alias({'a1': 'b2'}))
        self.assertEqual({'a1': 'b2', 'c3': 'd4'}, self.alias({'a1': 'b2', 'c3': 'd4'}))

    def test_big_collection(self):
        collection = [1, 2, [3, 4], 5, (6, 7), {8, 9}, [], 10, (), {}, {11: '12', '13': 14}]
        collection.append(collection[::])
        self.assertEqual(collection, self.alias(collection))

    def test_bytes(self):
        deserialized = self.alias(bts)
        self.assertEqual(bts, deserialized)
        self.assertEqual(bts.decode(), deserialized.decode())

    def test_bytearray(self):
        deserialized = self.alias(bts_arr)
        self.assertEqual(bts_arr, deserialized)
        self.assertEqual(bts_arr.decode(), deserialized.decode())


class FunctionsTestCase(BaseJsonTestCase):
    def test_using_modules(self):
        deserialized = self.alias(circle_area)
        self.assertEqual(circle_area(10), deserialized(10))

    def test_decorator(self):
        self.assertEqual(square_area(10), self.alias(square_area)(10))

        deserialized = self.alias(sum_func)
        self.assertEqual(sum_func(1, 2, 3, 4, 5), deserialized(1, 2, 3, 4, 5))
        try:
            deserialized(1, 2, 3, 4, 5, 6)
        except Exception as e:
            self.assertEqual(True, isinstance(e, ValueError))

    def test_closure(self):
        deserialized = self.alias(closure)
        self.assertEqual(type(closure), type(deserialized))
        self.assertEqual(closure(1, 2, 3)(), deserialized(1, 2, 3)())

    def test_lambda(self):
        deserialized = self.alias(lambda_pow)
        self.assertEqual(type(lambda_pow), type(deserialized))
        self.assertEqual(lambda_pow(36.6), deserialized(36.6))

    def test_recursion(self):
        deserialized = self.alias(factorial)
        self.assertEqual(type(factorial), type(deserialized))
        self.assertEqual(factorial(15), deserialized(15))

    def test_args_kwargs(self):
        self.assertEqual(sum_args(10, 20, 30), self.alias(sum_args(10, 20, 30)))

        self.assertEqual(sum_kwargs(a=10, b=20, c=30), self.alias(sum_kwargs(a=10, b=20, c=30)))

        self.assertEqual(sum_args_kwargs(1, 2, 3, a=1, b=2, c=3),
                         self.alias(sum_args_kwargs(1, 2, 3, a=1, b=2, c=3)))


class IterableTestCase(BaseJsonTestCase):
    def test_generator(self):
        deserialized = self.alias(generator())
        self.assertEqual(type(generator()), type(deserialized))
        self.assertEqual(sum(generator()), sum(deserialized))

    def test_generator_function(self):
        deserialized = self.alias(generator)
        self.assertEqual(type(generator()), type(deserialized()))
        self.assertEqual(sum(generator()), sum(deserialized()))

    def test_subgenerator(self):
        deserialized = self.alias(subgenerator())
        self.assertEqual(type(subgenerator()), type(deserialized))
        self.assertEqual(list(subgenerator()), list(deserialized))

    def test_generator_expression(self):
        deserialized = self.alias(generator_expression)
        self.assertEqual(type(generator_expression), type(deserialized))
        self.assertEqual(45, sum(deserialized))

    def test_iterator(self):
        deserialized = self.alias(it)
        self.assertEqual(45, sum(deserialized))


class ClassesTestCase(BaseJsonTestCase):
    def test_multiple_inheritance(self):
        deserialized = self.alias(C())
        self.assertEqual(str(type(C())), str(type(deserialized)))
        self.assertEqual(C().info_c(), deserialized.info_c())

    def test_long_inheritance(self):
        deserialized = self.alias(E())
        self.assertEqual(str(type(E())), str(type(deserialized)))
        self.assertEqual(E().info_e(), deserialized.info_e())

    def test_mro(self):
        self.assertEqual(str(C.__mro__), str(self.alias(C).__mro__))
        self.assertEqual(str(E.__mro__), str(self.alias(E).__mro__))

    def test_class_method(self):
        deserialized = self.alias(Human)
        #self.assertEqual(str(Human.get_const), str(deserialized.get_const))
        self.assertEqual(Human.get_const(), deserialized.get_const())

    def test_static_method(self):
        deserialized = self.alias(Human)
        self.assertEqual(type(Human.static), type(deserialized.static))
        self.assertEqual(Human.static(), deserialized.static())

    def test_property(self):
        deserialized = self.alias(Human)

        h1 = Human(18, 'Denis')
        h2 = deserialized(18, 'Denis')
        self.assertEqual(h1.age, h2.age)

        h1.age = 100
        h2.age = 100
        self.assertEqual(h1.age, h2.age)

        del h2.age
        self.assertEqual('name after age deletion', h2._name)
        self.assertEqual(False, hasattr(h2, 'age'))


class ObjectTestCase(BaseJsonTestCase):
    def test_class_object(self):
        attrs = {'age': 18,
                 'name': 'Pavel',
                 'email': 'qwe@asd.ru',
                 'phone': 123456}

        deserialized = self.alias(Profile(**attrs))
        self.assertEqual(str(Profile(**attrs)), str(deserialized))

    def test_property(self):

        human = Human(18, 'Denis')
        deserialized = self.alias(Human(18, 'Denis'))
        self.assertEqual(human.age, deserialized.age)

        human.age = 100
        deserialized.age = 100
        self.assertEqual(human.age, deserialized.age)

        del deserialized.age
        self.assertEqual('name after age deletion', deserialized._name)
        self.assertEqual(False, hasattr(deserialized, 'age'))


class ScopesTestCase(BaseJsonTestCase):
    def test_global(self):
        deserialized = self.alias(func1)
        self.assertEqual(10, deserialized())

    def test_nonlocal(self):
        deserialized = self.alias(func2)
        self.assertEqual(func2(), deserialized())

    def test_built_in(self):
        deserialized = self.alias(func3)
        self.assertEqual(func3([1, 2, 3]), deserialized([1, 2, 3]))

if __name__ == "__main__":
    unittest.main()


from serializers.base.constants import (PRIMITIVES, COLLECTIONS, CODE_ATTRIBUTES,
                                          TYPE, VALUE, METHOD_DECORATORS, ITERATOR_TYPE)
from types import NoneType, FunctionType, CodeType, CellType, MethodType
import inspect
import builtins


class Deserializer:

    @staticmethod
    def unpack(obj: dict):
        """
        Method that converts dictionary that
        represents the object to object.

        Parameters:
        obj (Any): Dictionary.

        Returns:
        object.
        """

        if obj[TYPE] in tuple(map(lambda p: p.__name__, PRIMITIVES)):
            if obj[TYPE] == str(NoneType.__name__):
                return None
            else:
                return getattr(builtins, obj[TYPE])(obj[VALUE])

        if obj[TYPE] in tuple(map(lambda c: c.__name__, COLLECTIONS)):
            if obj[TYPE] == dict.__name__:
                return {Deserializer.unpack(item[0]): Deserializer.unpack(item[1])
                        for item in obj[VALUE]}
            else:
                return getattr(builtins, obj[TYPE])(Deserializer.unpack(item)
                                                    for item in obj[VALUE])

        if obj[TYPE] in [FunctionType.__name__, MethodType.__name__]:
            return Deserializer._unpack_function(obj[VALUE])

        if obj[TYPE] == CodeType.__name__:
            return Deserializer._unpack_code(obj[VALUE])

        if obj[TYPE] == CellType.__name__:
            return CellType(Deserializer.unpack(obj[VALUE]))

        if obj[TYPE] == "class":
            return Deserializer._unpack_class(obj[VALUE])

        if obj[TYPE] in tuple(map(lambda md: md.__name__, METHOD_DECORATORS)):
            return getattr(builtins, obj[TYPE])(Deserializer.unpack(obj[VALUE]))

        if obj[TYPE] == property.__name__:
            return property(fget=Deserializer.unpack(obj[VALUE]["fget"]),
                            fset=Deserializer.unpack(obj[VALUE]["fset"]),
                            fdel=Deserializer.unpack(obj[VALUE]["fdel"]))

        if obj[TYPE] == ITERATOR_TYPE:
            return iter(Deserializer.unpack(item) for item in obj[VALUE])

        return Deserializer._unpack_object(obj[VALUE])

    @staticmethod
    def _unpack_function(obj: dict):
        code = Deserializer._unpack_code(obj["__code__"])

        globs = Deserializer._unpack_globals(obj["__globals__"], obj)
        globs["builtins"] = __import__("builtins")

        closure = Deserializer.unpack(obj["__closure__"])
        closure = tuple(closure) if closure else tuple()

        unpacked = FunctionType(code=code, globals=globs, closure=closure)
        unpacked.__globals__.update({unpacked.__name__: unpacked})
        unpacked.__defaults__ = Deserializer.unpack(obj["__defaults__"])
        unpacked.__kwdefaults__ = Deserializer.unpack(obj["__kwdefaults__"])

        return unpacked

    @staticmethod
    def _unpack_globals(globs, func):
        unpacked = dict()

        for key, value in globs.items():
            if "module" in key:
                unpacked[value[VALUE]] = __import__(value[VALUE])

            elif value != func["__name__"]:
                unpacked[key] = Deserializer.unpack(value)

        return unpacked

    @staticmethod
    def _unpack_class(obj):

        attrs = {member: Deserializer.unpack(value)
                 for member, value in obj.items()}

        cls = type(Deserializer.unpack(obj["__name__"]),
                   Deserializer.unpack(obj["__bases__"]),
                   attrs)

        for value in attrs.values():
            if inspect.isfunction(value):
                value.__globals__.update({cls.__name__: cls})
            elif isinstance(value, (staticmethod, classmethod)):
                value.__func__.__globals__.update({cls.__name__: cls})

        return cls

    @staticmethod
    def _unpack_object(obj):

        unpacked = object.__new__(Deserializer.unpack(obj["__class__"]))
        unpacked.__dict__ = {key: Deserializer.unpack(value)
                             for key, value in obj["__vars__"].items()}

        return unpacked

    @staticmethod
    def _unpack_code(code):
        # return CodeType(*(Deserializer.unpack(code[ATTRIBUTE])
        #                   for ATTRIBUTE in CODE_ATTRIBUTES))

        return CodeType(Deserializer.unpack(code["co_argcount"]),
                        Deserializer.unpack(code["co_posonlyargcount"]),
                        Deserializer.unpack(code["co_kwonlyargcount"]),
                        Deserializer.unpack(code["co_nlocals"]),
                        Deserializer.unpack(code["co_stacksize"]),
                        Deserializer.unpack(code["co_flags"]),
                        Deserializer.unpack(code["co_code"]),
                        Deserializer.unpack(code["co_consts"]),
                        Deserializer.unpack(code["co_names"]),
                        Deserializer.unpack(code["co_varnames"]),
                        Deserializer.unpack(code["co_filename"]),
                        Deserializer.unpack(code["co_name"]),
                        #Deserializer.unpack(code["co_qualname"]),
                        Deserializer.unpack(code["co_firstlineno"]),
                        Deserializer.unpack(code["co_lnotab"]),
                        #Deserializer.unpack(code["co_exceptiontable"]),
                        Deserializer.unpack(code["co_freevars"]),
                        Deserializer.unpack(code["co_cellvars"]))


from serializers.json_serializer.JsonSerializer import JsonSerializer
from serializers.xml_serializer.XmlSerializer import XmlSerializer


class SerializersFactory:
    @staticmethod
    def create_serializer(name: str):
        match name.lower():
            case "json":
                return JsonSerializer
            case "xml":
                return XmlSerializer
            case _:
                raise ValueError(f"{name} is not supported!")



def serialize_to_xml(obj) -> str:
    if type(obj) == tuple:
        serialized = []
        for i in obj:
            if type(i) != tuple:
                serialized.append(
                    f"<str>{serialize_to_xml(i)}</str> ")
            else:
                serialized.append(
                    f"{serialize_to_xml(i)}")
        ans = "".join(serialized)
        return f"<{type(obj).__name__}>{ans}</{type(obj).__name__}> "
    else:
        return obj


def deserialize_xml(obj: str):
    if obj == '<tuple></tuple>':
        return tuple()
    elif obj[:7] == '<tuple>':
        obj = obj[7:-9]
        if obj[-1] == ' ':
            obj = obj[:-1]

        parsed = []
        depth = 0
        substr = ""
        for i in obj:
            if i == '<' or i == '>':
                depth += 1
            elif i == '/':
                depth -= 4
            elif depth == 0:
                parsed.append(deserialize_xml(substr))
                substr = ""
                continue

            substr += i
        parsed.append(deserialize_xml(substr))
        return tuple(parsed)

    elif obj[:5] == '<str>':
        parsed = []
        ind = obj.find('</str>')
        if obj[ind + 6:] != "":
            parsed.append(deserialize_xml(obj[5:ind]))
            parsed.append(deserialize_xml(obj[ind + 6:]))
        else:
            return obj[5:ind]
        return tuple(parsed)
    else:
        return obj



from typing import Any
from serializers.base.deserialization import Deserializer
from serializers.base.serialization import Serializer
from ..base.constants import (PRIMITIVES,
                         JsonRegularExpression as Expression)
import regex


class JsonSerializer:
    def dumps(self, obj: Any) -> str:
        """
        Convert an object to string of 'JSON' format.

        :param obj: Any python object.
        :type obj: Any

        :return: String of 'JSON' format.
        :rtype: str
        """

        packed: dict = Serializer.pack(obj)
        return self._convert(packed)

    def loads(self, string: str):
        """
        Convert string of 'JSON' format to object.

        :param string: String of 'JSON' format.
        :type string: str

        :return: Python object.
        :rtype: Any
        """

        packed: dict = self._deconvert(string)
        return Deserializer.unpack(packed)

    def _convert(self, obj: dict) -> str:
        """
        Convert dictionary that represents object
        to a string of 'JSON' format

        :param obj: Dictionary that represents object.
        :type obj: dict

        :return: String of 'JSON' format.
        :rtype: str
        """

        if isinstance(obj, PRIMITIVES):
            if isinstance(obj, str):
                return '"' + obj.replace("\\", "\\\\").replace('"', "\"").replace("'", "\'") + '"'
            else:
                return str(obj)

        if isinstance(obj, list):
            return '[' + ', '.join([self._convert(item) for item in obj]) + ']'

        if isinstance(obj, dict):
            return '{' + ', '.join([f'{self._convert(key)}: {self._convert(value)}'
                                    for key, value in obj.items()]) + '}'

    def _deconvert(self, string: str) -> dict:
        """
        Convert string of 'JSON' format to
        dictionary that represents object.

        :param string: String of 'JSON' format.
        :type string: str

        :return: Dictionary that represents object.
        :rtype: dict
        """

        string = string.strip()

        if regex.fullmatch(Expression.INT.value, string):
            return int(string)

        if regex.fullmatch(Expression.STR.value, string):
            string = string.replace("\\\\", "\\").replace(r"\"", '"').replace(r"\'", "'")
            return string[1:-1]

        if regex.fullmatch(Expression.FLOAT.value, string):
            return float(string)

        if regex.fullmatch(Expression.BOOL.value, string):
            return True if string == 'True' else False

        if regex.fullmatch(Expression.NONE.value, string):
            return None

        if string.startswith("[") and string.endswith("]"):
            string = string[1:-1]
            matches = regex.findall(Expression.ANY_VALUE.value, string)
            return [self._deconvert(match[0]) for match in matches]

        if string.startswith("{") and string.endswith("}"):
            string = string[1:-1]
            matches = regex.findall(Expression.ANY_VALUE.value, string)
            return {self._deconvert(matches[i][0]): self._deconvert(matches[i + 1][0])
                    for i in range(0, len(matches), 2)}



def serialize_to_json(obj) -> str:
    """ 
    Function:
    -----------
    Serializes object to josn string

    Parameters:
    -----------
           - obj: object to serialize

    Returns:
    -----------
          - string
    """
    if type(obj) == tuple:
        serialized = []
        for i in obj:
            serialized.append(f"{serialize_to_json(i)}")
        ans = ", ".join(serialized)
        return f"[{ans}]"
    else:
        return f"\"{str(obj)}\""


def deserialize_json(obj: str):
    if obj == '[]':
        return tuple()
    elif obj[0] == '[':
        obj = obj[1:len(obj) - 1]
        deserialized_obj = []
        depth = 0
        is_quote = False
        substr = ""
        for i in obj:
            if i == '[':
                depth += 1
            elif i == ']':
                depth -= 1
            elif i == '\"':
                is_quote = not is_quote
            elif i == ',' and not is_quote and depth == 0:
                deserialized_obj.append(deserialize_json(substr))
                substr = ""
                continue
            elif i == ' ' and not is_quote:
                continue

            substr += i

        deserialized_obj.append(deserialize_json(substr))
        return tuple(deserialized_obj)
    else:
        return obj[1:len(obj) - 1]



from typing import Final
from types import NoneType
import types
import enum


def is_iterable(obj):
    return hasattr(obj, '__iter__') and hasattr(obj, '__next__') and callable(obj.__iter__)

XML: Final[str] = "xml"
JSON: Final[str] = "json"

KEY: Final[str] = "key"
TYPE: Final[str] = "type"
VALUE: Final[str] = "value"

PRIMITIVES: Final[list] = (int, float, bool, str, NoneType, complex)
COLLECTIONS: Final[list] = (set, dict, list, tuple, bytes, bytearray)

CODE_ATTRIBUTES: Final[list] = ("co_argcount",
                                "co_posonlyargcount",
                                "co_kwonlyargcount",
                                "co_nlocals",
                                "co_stacksize",
                                "co_flags",
                                "co_code",
                                "co_consts",
                                "co_names",
                                "co_varnames",
                                "co_filename",
                                "co_name",
                                "co_qualname",
                                "co_firstlineno",
                                "co_lnotab",
                                "co_exceptiontable",
                                "co_freevars",
                                "co_cellvars")

IGNORED_CLASS_ATTRIBUTES: list[str] = ("__name__",
                                       "__base__",
                                       "__basicsize__",
                                       "__dictoffset__",
                                       "__class__")

IGNORED_TYPES: Final[list] = (types.WrapperDescriptorType,
                              types.MethodDescriptorType,
                              types.BuiltinFunctionType,
                              types.GetSetDescriptorType,
                              types.MappingProxyType)

METHOD_DECORATORS: Final[list] = (classmethod, staticmethod)

ITERATOR_TYPE: Final[str] = "iterator"


class JsonRegularExpression(enum.Enum):
    INT = r"[+-]?\d+"
    FLOAT = fr"({INT}(?:\.\d+)?(?:e{INT})?)"
    BOOL = r"((True)|(False))\b"
    STR = r"\"((\\\")|[^\"])*\""
    NONE = r"\b(None)\b"
    COMPLEX = fr"{FLOAT}{FLOAT}j"

    LIST_RECURSION = r"\[(?R)?(,(?R))*\]"
    DICT_RECURSION = r"\{((?R):(?R))?(?:,(?R):(?R))*\}"

    ANY_VALUE = fr"\s*({LIST_RECURSION}|" \
                fr"{DICT_RECURSION}|" \
                fr"{STR}|" \
                fr"{FLOAT}|" \
                fr"{BOOL}|" \
                fr"{INT}|" \
                fr"{NONE}|" \
                fr"{COMPLEX}\s*)"


class XmlRegularExpression(enum.Enum):
    TYPES = "|".join(map(lambda x: x.__name__, list(PRIMITIVES) + [list, dict]))
    ITEM = fr"\s*(\<(?P<{KEY}>{TYPES})\>(?P<{VALUE}>([^<>]*)|(?R)+)\</({TYPES})\>)\s*"


from setuptools import setup


setup(
    name="Shchirov_serializer",
    version="1.1.8",
    description="python library for serialization",
    url="https://github.com/PaShampusik/PythonLabs/Lab_3",
    author="PaShampusik",
    author_email="shchirovpavel@gmail.com",
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent"
    ],
    install_requires = [
        'regex'
    ],
    packages=["serializers/json_serializer", "serializers/base",
              "serializers/xml_serializer", "serializers"],
    include_package_data=True
)



from typing import Any
from serializers.base.deserialization import Deserializer
from serializers.base.serialization import Serializer
from types import NoneType
from ..base.constants import (PRIMITIVES, KEY, VALUE,
                         XmlRegularExpression as Expression)
import regex
import builtins


class XmlSerializer():

    def dumps(self, obj: Any) -> str:
        """
        Convert an object to string of 'XML' format.

        :param obj: Any python object.
        :type obj: Any

        :return: String of 'XML' format.
        :rtype: str
        """

        packed: dict = Serializer.pack(obj)
        return self._convert(packed)

    def loads(self, string: str):
        """
        Convert string of 'XML' format to object.

        :param string: String of 'XML' format.
        :type string: str

        :return: Python object.
        :rtype: Any
        """

        packed: dict = self._deconvert(string)
        return Deserializer.unpack(packed)

    def _convert(self, obj: dict) -> str:
        """
        Convert dictionary that represents object
        to a string of 'XML' format

        :param obj: Dictionary that represents object.
        :type obj: dict

        :return: String of 'XML' format.
        :rtype: str
        """

        if isinstance(obj, PRIMITIVES):
            if isinstance(obj, str):
                return XmlSerializer.create_xml_item(type(obj).__name__, XmlSerializer.to_special_xml(obj))
            else:
                return XmlSerializer.create_xml_item(type(obj).__name__, str(obj))

        if isinstance(obj, list):
            return XmlSerializer.create_xml_item(type(obj).__name__, "".join([self._convert(item) for item in obj]))

        if isinstance(obj, dict):
            return XmlSerializer.create_xml_item(type(obj).__name__, "".join(
                [f"{self._convert(key)}{self._convert(value)}" for key, value in obj.items()]))

    def _deconvert(self, string: str) -> dict:
        """
        Convert string of 'XML' format to
        dictionary that represents object.

        :param string: String of 'XML' format.
        :type string: str

        :return: Dictionary that represents object.
        :rtype: dict
        """

        string = string.strip()

        match = regex.fullmatch(Expression.ITEM.value, string)
        if not match:
            return

        key = match.group(KEY)
        value = match.group(VALUE)

        if key in map(lambda p: p.__name__, PRIMITIVES):
            if key == str.__name__:
                return XmlSerializer.from_special_xml(value)
            elif key == NoneType.__name__:
                return None
            elif key == bool.__name__:
                return True if value == 'True' else False
            else:
                return getattr(builtins, key)(value)

        if key == list.__name__:
            matches = regex.findall(Expression.ITEM.value, value)
            return [self._deconvert(match[0]) for match in matches]

        if key == dict.__name__:
            matches = regex.findall(Expression.ITEM.value, value)
            return {self._deconvert(matches[i][0]): self._deconvert(matches[i + 1][0])
                    for i in range(0, len(matches), 2)}

    @staticmethod
    def create_xml_item(tag_name, value):
        return f"<{tag_name}>{value}</{tag_name}>"

    @staticmethod
    def to_special_xml(string):
        return string.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;"). \
            replace('"', "&quot;").replace("'", "&apos;")

    @staticmethod
    def from_special_xml(string):
        return string.replace("&amp;", "&").replace("&lt;", "<").replace("&gt;", ">"). \
            replace("&quot;", '"').replace("&apos;", "'")